package cz.upol.jj2.lab03;

import java.util.Arrays;
import java.util.concurrent.*;


/**
 * Ukazky paralelismu na vyssi urovni (Executor)
 * Priklad: Fibonacciho cisla (rekurzivne, naivne, bez optimalizaci)
 * 1. pocita prvnich n Fibonacciho cisel
 *   - jednou sekvencne - jedno po druhem
 *   - jednou paralelne - kazde v jednom vlakne
 * 2. pocita n-te Fibonacciho cislo
 *   - jednou sekvencne
 *   - jednou paralelne - pomoci Fork/Join frameworku
 */
public class Fibonacci {
      /**
     * misto pro vysledky
     */
    static long[] results;

    /**
     * pocet cisel ke spocitani
     */
    static int n;

    public static void main(String[] args) throws Exception {
        // prnvich n Fibonacciho cisel sekvencne/paralalne
        //kolik fibonacciho cisel chceme?
        n = 40;
        results = new long[n];
        int threadCount = 8; //zkuste ruzne pocty vlaken

        //sekvenci vypocet zacatku Fibnacciho posloupnosti
        long start = System.currentTimeMillis(); //vrati aktualni cas v ms
        singleThreadFibonacciSequence();
        long end = System.currentTimeMillis();
        System.out.println(String.format("%d Fibonacciho cisel spocitano v case %d ms.", n, end - start));  //rozdilem end - start dostaneme kolik ms ubehlo
        System.out.println(Arrays.toString(results));

        //paralelni vypocet zacatku Fibonacciho posloupnosti
        start = System.currentTimeMillis();
        multiThreadFibonacciSequence(threadCount);
        end = System.currentTimeMillis();
        System.out.println(String.format("%d Fibonacciho cisel paralelne spocitano v case %d ms pri %d vlaknech.", n, end - start, threadCount));
        System.out.println(Arrays.toString(results));

        //---------------------------------------------
        // jedno Fibonacciho cislo sekvence/paralelne
        n = 45;
        threadCount = 8;

        //sekvenci n-te Fib cislo
        start = System.currentTimeMillis();
        long result = singleThreadFibonacciNumber(n);
        end = System.currentTimeMillis();
        System.out.println(String.format("%d-te Fibonacciho cislo (%d) spocitano v case %d ms.", n, result, end - start));

        //paralelni n-te Fibonacciho cislo  pomocí Fork/Join framewroku pro efektivni implementaci rekurzivnich algoritmu
        start = System.currentTimeMillis();
        result = multiThreadFibonacciNumber(n, threadCount);
        end = System.currentTimeMillis();
        System.out.println(String.format("%d-te Fibonacciho cislo (%d) spocitano v case %d ms pri %d vlaknech.", n, result, end - start, threadCount));
    }

    /**
     * Klasicky rekurzivni vypocet n-teho Fibonacciho cisla
     * @param n
     * @return
     */
    public static long singleThreadFibonacciNumber(int n) {
        if (n <= 1) {
            return 1;
        }
        return singleThreadFibonacciNumber(n - 1) + singleThreadFibonacciNumber(n - 2);
    }


    /**
     * Prvnich n Fibonacciho cisel do pripraveneho pole sekvencne
     */
    public static void singleThreadFibonacciSequence() {
        for (int i = 0; i < n; i++) {
            results[i] = singleThreadFibonacciNumber(i);
        }
    }

    /**
     * Prvni n Fibonacciho cisel do pripravenho pole paralelne (pomoci klasického poolu vláken)
     * @param threadCount
     * @throws Exception
     */
    public static void multiThreadFibonacciSequence(int threadCount) throws Exception {
        //pool vlaken
        ExecutorService executors = Executors.newFixedThreadPool(threadCount);
        //misto na prisliby vysledku
        Future<Long>[] futureResults = new Future[n];

        for (int i = 0; i < n; i++) {
            final int constI = i; // do Callable muze jen konstatni obsah
            //vytvor prisliby
            futureResults[i] = executors.submit(() -> singleThreadFibonacciNumber(constI)); //insatnce Callable vyjadrena lambdou
        }

        for (int i = 0; i < n; i++) {

            try {
                //vyzvedni hodnoty prislibu
                results[i] = futureResults[i].get();
            } catch (InterruptedException | ExecutionException e) {
                e.printStackTrace();
                throw new Exception("Problem s vlakny", e);
            }
        }

        //zrusit vytvorena vlakna
        executors.shutdown();
    }

    /**
     * N-te Fibonacciho číslo paralelně pomoci fork join frameworku
     * @param n
     * @param threadCount
     * @return
     */
    public static long multiThreadFibonacciNumber(int n, int threadCount) {
        //pool vlaken
        ForkJoinPool pool = new ForkJoinPool(threadCount);
        //vypovcet popsany pomocí RecursiveTask
        FibonacciTask task = new FibonacciTask(n);
        return pool.invoke(task);
    }

    /**
     * Task pro pocitani n-teho fibonacciho cisla
     */
    private static class FibonacciTask extends RecursiveTask<Long> {

        /**
         * Kolikate cislo budeme pocitat
         */
        private final int n;

        public FibonacciTask(int n) {
            this.n = n;
        }

        @Override
        protected Long compute() {
            if (n <= 30) {      //pro mala cisla by byla rezie neunosna - zkuste!
                return singleThreadFibonacciNumber(n);
            } else {
                FibonacciTask f1 = new FibonacciTask(n - 1);
                FibonacciTask f2 = new FibonacciTask(n - 2);
                f1.fork();  //f1 v dalsim vlakne (vlakno doda pool, az bude nejake volne)
                long result2 = f2.compute(); //toto vlakno pokracuje vypotem f2
                long result1 = f1.join(); // pockame na f1, je-li to treba
                return result1 + result2;
            }
        }
    }

}
